;//---------------------------------------------------------------------
;// Copyright (C)2002 Matt J. Scott.
;//
;// This file is subject to the terms and conditions of the GNU General
;// Public License Version 2.  See the file "COPYING" in the main
;// directory of this archive for more details
;//---------------------------------------------------------------------

.init_vf_all	; allow VCL to use all VF regs
.init_vi_all	; allow VCL to use all VI regs
.syntax new

;;;
;;; CONSTANTS
;;;
	
; pad const
NOPAD 			.assign		32767		; 0x7FFF

START			.assign		8			; 0x0008
UP				.assign		16			; 0x0010
RIGHT			.assign		32			; 0x0020
DOWN			.assign		64			; 0x0040
LEFT			.assign		128			; 0x0080
TRIANGLE		.assign		4096		; 0x1000
CIRCLE			.assign		8192		; 0x2000
X				.assign		16384		; 0x4000
SQUARE			.assign		32767		; 0x7FFF
R2				.assign		512			; 0x0200
R1				.assign		2048		; 0x0800

; helper consts
AllZeroMACFlag	.assign		14			; 0x000E
AllOneMask		.assign		32767		; 0x7FFF
OnlyMoveMask	.assign		32527		; 0x7F0F
idletime		.assign		300
resetauto		.assign		30720		; 0x7800
autosamplerate1	.assign		31			; 0x001F
autosamplerate2	.assign		16			; 0x0010

; data constants
;
NumParticles	.assign		53			; number of particles
NumSticks		.assign		60			; number of stick constraints
NumMeshs		.assign		10			; number of total meshes
NumLimbs		.assign		8			; number of limb meshes
NumBodys		.assign		2			; number of body meshes
	
;;;
;;; MEMORY LAYOUT
;;;

; control info
ControlQW		.assign		0			; Control QuadWord

; preformatted giftags and rgbaq packets
colorGifTag		.assign		1			; gif tag for list of sprites
sprGifTag		.assign		2			; gif tag for list of sprites
lineGifTag		.assign		3			; gif tag for list of lines
meshGifTag		.assign		4			; gif tag for meshes
linesGifTag		.assign		5			; gif tag for line strips
SyncGifTag		.assign		6			; gif tag for Syncing
fgrey200		.assign		7			; RGBAQ REG for grey200
fgrey128		.assign		8			; RGBAQ REG for grey128
green			.assign		9			; RGBAQ REG for green
grey			.assign		10			; RGBAQ REG for green
	
; Some nice constants and static matrix
GravityConst	.assign		11			; gravity
MinimumVecBig	.assign		12
MaximumVecBig	.assign		13	
MinimumVecSmall	.assign		14
MaximumVecSmall	.assign		15	

ViewScreen0		.assign		16			; camera -> screen
ViewScreen1		.assign		17			;
ViewScreen2		.assign		18			;
ViewScreen3		.assign		19			;

LightVector0	.assign		20			; light 0 vector
LightVector1	.assign		21			; light 1 vector
LightVector2	.assign		22			; light 2 vector
LightVector3	.assign		23			; light 3 vector

LightColor0		.assign		24			; light 0 vector
LightColor1		.assign		25			; light 1 vector
LightColor2		.assign		26			; light 2 vector
LightColor3		.assign		27			; light 3 vector

; initial camera pos
WorldView0		.assign		28			; world -> view
WorldView1		.assign		29			;
WorldView2		.assign		30			;
WorldView3		.assign		31			;

; initial indent matrix for local to world calcs
LocalWorld0		.assign		32			; local -> world
LocalWorld1		.assign		33			; local -> world
LocalWorld2		.assign		34			; local -> world
LocalWorld3		.assign		35			; local -> world

; data 
ParticlePos		.assign		36			; Particle postition
ParticlePrevPos	.assign		89			; Particle Postition last frame
StickList		.assign		142			; Start of the 59 sticks constraints
MeshList		.assign		202			; Start of the 12 mesh instance inro
SaveCtlPos		.assign		226			; 5 qw to save control points
OrigCtlPos		.assign		231			; 5 qw to save orig control points
LimbMeshData	.assign		236			; LimbMesh Data
BodyMeshData	.assign		726			; BodyMesh Data
LineStripData	.assign		870			; line strip data
OutputPacket	.assign		895			; Start of output packet
LastQW			.assign		1023		; end of memory

;;;
;;; THE PROGRAM
;;;

--enter
--endenter
	
; setup code - only need to execute once
;
	lq			v2s0, ViewScreen0(vi00)			; load view to screen matrix
	lq			v2s1, ViewScreen1(vi00)			; 
	lq			v2s2, ViewScreen2(vi00)			; 
	lq			v2s3, ViewScreen3(vi00)			; 
	
	lq			nl0, LightVector0(vi00)			; load the light normals
	lq			nl1, LightVector1(vi00)			;
	lq			nl2, LightVector2(vi00)			;
	lq			nl3, LightVector3(vi00)			;
	
	lq			lCol0, LightColor0(vi00)		; load the light colors
	lq			lCol1, LightColor1(vi00)		;
	lq			lCol2, LightColor2(vi00)		;
	lq			lCol3, LightColor3(vi00)		;
	
	lq			w2v0, WorldView0(vi00)			; load world to view matrix
	lq			w2v1, WorldView1(vi00)			; 
	lq			w2v2, WorldView2(vi00)			; 
	lq			w2v3, WorldView3(vi00)			; 

	ilw.z		matrixmode, ControlQW(vi00)z	; load the matrix mode
	ibeq		matrixmode, vi00, nomatrixcam	; no need to do cam update
	ilw.w		mmcnt, ControlQW(vi00)w			; how long to stay in matrixmode
	isubiu		mmcnt, mmcnt, 1
	isw.w		mmcnt, ControlQW(vi00)w			; save the mmcnt
	ibeq		mmcnt, vi00, mmdone

	sub.y		rot0, vf00, vf00				; clear
	sub.xz		rot1, vf00, vf00				; clear
	sub.y		rot2, vf00, vf00				; clear
	addw.y		rot1, vf00, vf00				; 0.0   1.0 0.0
	loi			0.99862953
	addi.x		rot0, vf00, i					; cos() 0.0 0.0
	addi.z		rot2, vf00, i					; 0.0   0.0 cos()
	loi			0.05233595
	addi.z		rot0, vf00, i					; cos() 0.0 sin()
	loi			-0.05233595
	addi.x		rot2, vf00, i					; -sin() 0.0 cos()

	mulax.xyz	acc,  rot0, w2v0
	madday.xyz	acc,  rot1, w2v0
	maddz.xyz	w2v0, rot2, w2v0
	mulax.xyz	acc,  rot0, w2v1
	madday.xyz	acc,  rot1, w2v1
	maddz.xyz	w2v1, rot2, w2v1
	mulax.xyz	acc,  rot0, w2v2
	madday.xyz	acc,  rot1, w2v2
	maddz.xyz	w2v2, rot2, w2v2

	loi			120.00
	addi.z		w2v3, vf00, i					; move back a bit

	sq.xyz		w2v0, WorldView0(vi00)
	sq.xyz		w2v1, WorldView1(vi00)
	sq.xyz		w2v2, WorldView2(vi00)
	sq.z		w2v3, WorldView3(vi00)
	b 			nomatrixcam

	mmdone:
	iaddiu		matrixmode, vi00, 0				; clear the matrix mode
	isw.z		matrixmode, ControlQW(vi00)z	; 

	sub.yz		w2v0, vf00, vf00				; clear
	addw.x		w2v0, vf00, vf00				; 1.0 0.0 0.0
	sq.xyz		w2v0, WorldView0(vi00)

	sub.xz		w2v1, vf00, vf00				; clear
	addw.y		w2v1, vf00, vf00				; 0.0 1.0 0.0
	sq.xyz		w2v1, WorldView1(vi00)

	sub.xy		w2v2, vf00, vf00				; clear
	addw.z		w2v2, vf00, vf00				; 0.0 0.0 1.0
	sq.xyz		w2v2, WorldView2(vi00)

	loi			100.00
	addi.z		w2v3, vf00, i					; move back a bit
	sq.z		w2v3, WorldView3(vi00)

nomatrixcam:
	; [World->Screen] = [World->View][View->Screen]
	mulax		acc,  v2s0, w2v0
	madday		acc,  v2s1, w2v0
	maddaz		acc,  v2s2, w2v0
	maddw		w2s0, v2s3, w2v0
	mulax		acc,  v2s0, w2v1
	madday		acc,  v2s1, w2v1
	maddaz		acc,  v2s2, w2v1
	maddw		w2s1, v2s3, w2v1
	mulax		acc,  v2s0, w2v2
	madday		acc,  v2s1, w2v2
	maddaz		acc,  v2s2, w2v2
	maddw		w2s2, v2s3, w2v2
	mulax		acc,  v2s0, w2v3
	madday		acc,  v2s1, w2v3
	maddaz		acc,  v2s2, w2v3
	maddw		w2s3, v2s3, w2v3
	
	; skip all processing if in MATRIX mode	
	ilw.z		matrixmode, ControlQW(vi00)z	; load the matrix mode
	ibne		matrixmode, vi00, onlyrender

; Accumulate forces and Verlet integration
;
	iaddiu		pPos, vi00, ParticlePos			; pointer to particles pos
	iaddiu		pPrevPos, vi00, ParticlePrevPos ; pointer to particles prev pos
	iaddiu		nParticles, vi00, NumParticles	; number of particles
	lq 			gravity, GravityConst(vi00)		; set gravity const
	loi 		2.0								; load up 2.0 const
	
verlet_loop:
	
	lq.xyzw		pos, (pPos)						; load up the particle pos
	lq.xyzw		prevpos, (pPrevPos)				; load up the particle prev pos
	sqi.xywz	pos, (pPrevPos++)				; update prev pos

	; accumulate forces
	mulw 		accel, gravity, pos				; accel = gravity*inverse mass
												; w component of pos is invmass

	; Verlet integration						; pos'=((pos*2)-prevpos)+accel
	muli		pos, pos, I						; pos * 2.0f
	sub			pos, pos, prevpos				; pos - prevpos
	add			pos, pos, accel					; pos + accel
	sqi.xywz	pos, (pPos++)					; update pos

	isubiu		nParticles, nParticles, 1		; dec the # of particles
	ibne		nParticles, vi00, verlet_loop
	

; process pad input to adjust control points pos
;

	iaddiu		movex, vi00, 0					; initialize 0 x movement
	iaddiu		movey, vi00, 0					; initialize 0 y movement
	iaddiu		movez, vi00, 0					; initialize 0 z movement
	iaddiu		autopilot, vi00, 0				; not in autopilot mode

	iaddiu		none, vi00, NOPAD
	iaddi		highbit, none, 1
	iadd		none, highbit, none				; Make an all ones mask

	ilw.w		pad, LastQW(vi00)				
	ilw.y		prevpad, ControlQW(vi00)y		; load the previous frame pad

	iand		input, pad, none
	ibne		input, none, gotinput

; If we don't get input for awhile, then make the marionett do some thing
;
	ilw.x		noinputcnt, ControlQW(vi00)		; load up the noinput cnt
	iaddiu		mask, vi00, resetauto			; when to roll over counter
	iand		reset, noinputcnt, mask
	ibeq		reset, vi00, noresetcnt
	iaddiu		noinputcnt, vi00, 0				; reset the autocnt
	isw.x		noinputcnt, ControlQW(vi00)		; store the autocnt

	noresetcnt:
	iaddiu		noinputcnt, noinputcnt, 1		; inc the cnt
	isw.x		noinputcnt, ControlQW(vi00)		; store the noinput cnt
	isubiu		length, noinputcnt, idletime
	ibltz		length, handleinput				; not long engough

	ibgtz		length, autopilot				; auto movement
	
	iaddiu		iter, vi00, 5					; for each control point move it
	iaddiu		pPos, vi00, ParticlePos			; pointer to list of particles
	iaddiu		pPosPrev, vi00, ParticlePrevPos	; pointer to particle prev pos
	iaddiu		pOrigCtlPos, vi00, OrigCtlPos	; pointer to mem to orig ctlpts

	restoreorigctlpts_loop:
	lqi.xyz		pos, (pOrigCtlPos++)			; load up the orig particle pos
	sqi.xyz		pos, (pPos++)					; reset the pos of the ctl pts
	sqi.xyz		pos, (pPosPrev++)				; reset the pos of the ctl pts

	isubiu		iter, iter, 1					; dec the # of iterations
	ibne		iter, vi00, restoreorigctlpts_loop

	autopilot:
	iaddiu		mask1, vi00, autosamplerate1	; how often to sample input
	iaddiu		mask2, vi00, autosamplerate2	; 
	iand		read, noinputcnt, mask1
	isub		read, read, mask2
	ibne		read, vi00, noreadauto
	ilw.z		pad, LastQW(vi00)
	;lq.z		fpad, LastQW(vi00)				; load pad from the rand numb
	;ftoi4.z		fpad, fpad
	;mtir		pad, fpad[z]
	b			processauto
	noreadauto:
	lq.w		fpad, LastQW(vi00)				; load pad from prev pad
	ftoi4.w		fpad, fpad
	mtir		pad, fpad[w]

	processauto:
	iadd		pad, pad, noinputcnt
	iaddiu		mask1, vi00, AllOneMask			; 0x7FFF
	iaddiu		mask2, mask1, 1					; 0x8000
	iaddiu		mask1, vi00, OnlyMoveMask		; 0x700F
	iadd		mask1, mask1, mask2				; 0xF00F
	ior			pad, pad, mask1
	isw.w		pad, LastQW(vi00)				; 
	iaddiu		autopilot, vi00, 1				; remember that we in autopilot
	b			handleinput

gotinput:
	isw.x		vi00, ControlQW(vi00)			; clear the noinput cnt

handleinput:
	iaddiu		button, vi00, START
	iand		on, pad, button					; look for start
	ibne		on, vi00, nostart
	iand		prevon, prevpad, button			; look for x last frame
	ibeq		prevon, vi00, nostart
	iaddiu		domatrixmode, vi00, 1			; set the matrix mode flag
	isw.z		domatrixmode, ControlQW(vi00)z	; store the matrix mode flag
	iaddiu		domatrixmode, vi00, 120			; 120 frames
	isw.w		domatrixmode, ControlQW(vi00)w	; init the mmcnt to 1

	nostart:
	iaddiu		button, vi00, R2
	iand		on, pad, button					; look for R2
	ibne		on, vi00, nor2
	iaddi		movez, movez, 1					; move 1 x
	;b 			nor1

	nor2:
	iaddiu		button, vi00, R1
	iand		on, pad, button					; look for R1
	ibne		on, vi00, nor1
	iaddi		movez, movez, -1				; move -1 x

	nor1:
	iaddiu		button, vi00, LEFT
	iand		on, pad, button					; look for dpad left
	ibne		on, vi00, noleft
	iaddi		movex, movex, -1				; move 1 x
	;b 			noright

	noleft:
	iaddiu		button, vi00, RIGHT
	iand		on, pad, button					; look for dpad right
	ibne		on, vi00, noright
	iaddi		movex, movex, 1					; move -1 x

	noright:
	iaddiu		button, vi00, UP
	iand		on, pad, button					; look for dpad up
	ibne		on, vi00, noup
	iaddi		movey, movey, -1				; move 1 y
	;b 			nodown

	noup:
	iaddiu		button, vi00, DOWN
	iand		on, pad, button					; look for dpad down
	ibne		on, vi00, nodown
	iaddi		movey, movey, 1					; move -1 y

	nodown:
	iaddiu		tug0, vi00, 0					; null out tugs
	iaddiu		tug2, vi00, 0					; 
	iaddiu		tug3, vi00, 0					; 
	iaddiu		tug4, vi00, 0					; 

	iaddiu		button, vi00, X
	iand		on, pad, button					; look for x
	iand		prevon, prevpad, button			; look for x last frame
	ibne		prevon, vi00, noxlasttime
	ibeq		on, vi00, donex 
	iaddi		tug3, vi00, +4					; move +4 y on string 3
	b			donex
	noxlasttime:
	ibne		on, vi00, donex 
	iaddi		tug3, vi00, -4					; move -4 y on string 3

	donex:
	iaddiu		button, vi00, CIRCLE
	iand		on, pad, button					; look for circle
	iand		prevon, prevpad, button			; look for x last frame
	ibne		prevon, vi00, nocirclelasttime
	ibeq		on, vi00, donecircle 
	iaddi		tug4, vi00, +4					; move +4 y on string 3
	b			donecircle
	nocirclelasttime:
	ibne		on, vi00, donecircle 
	iaddi		tug4, vi00, -4					; move -4 y on string 3

	donecircle:
	iaddiu		button, vi00, SQUARE
	iaddi		button, button, 1
	iand		on, pad, button				; look for 
	iand		prevon, prevpad, button			; look for x last frame
	ibne		prevon, vi00, nosquarelasttime
	ibeq		on, vi00, donesquare 
	iaddi		tug0, vi00, +10					; move +10 y on string 3
	b			donesquare
	nosquarelasttime:
	ibne		on, vi00, donesquare 
	iaddi		tug0, vi00, -10					; move -10 y on string 3

	donesquare:
	iaddiu		button, vi00, TRIANGLE
	iand		on, pad, button				; look for triangle
	iand		prevon, prevpad, button			; look for x last frame
	ibne		prevon, vi00, notrianglelasttime
	ibeq		on, vi00, donetriangle 
	iaddi		tug2, vi00, +10					; move +10 y on string 3
	b			donetriangle
	notrianglelasttime:
	ibne		on, vi00, donetriangle 
	iaddi		tug2, vi00, -10					; move -10 y on string 3

	donetriangle:

; move the control points
;
	iaddiu		iter, vi00, 5					; for each control point move it
	iaddiu		pPos, vi00, ParticlePos			; pointer to list of particles
	iaddiu		pPosPrev, vi00, ParticlePrevPos	; pointer to particle prev pos
	iaddiu		pSaveCtlPos, vi00, SaveCtlPos	; pointer to mem to save ctlpts

movectlpt_loop:

	lq.xyz		pos, (pPos)						; load up the particle pos
	sqi.xyz		pos, (pSaveCtlPos++)			; save the pos of the ctl pts

	mfir.x		fmovex, movex					; move to a fp reg
	itof0.x		fmovex, fmovex					; make int a float
	addx.x		pos, pos, fmovex				; add x motion

	mfir.y		fmovey, movey					; move to a fp reg
	itof0.y		fmovey, fmovey					; make int a float
	addy.y		pos, pos, fmovey				; add y motion

	mfir.z		fmovez, movez					; move to a fp reg
	itof0.z		fmovez, fmovez					; make int a float
	addz.z		pos, pos, fmovez				; add z motion

	sqi.xyz		pos, (pPos++)					; store the new pos
	sqi.xyz		pos, (pPosPrev++)				; store the new pos as prev pos

	isubiu		iter, iter, 1					; dec the # of particles
	ibne		iter, vi00, movectlpt_loop
	
; constrain the movement
;
	iaddiu		iter, vi00, 5					; for each control point move it
	iaddiu		pPos, vi00, ParticlePos			; pointer to list of particles
	iaddiu		allZeros, vi00, AllZeroMACFlag	; the last op set all vec to 0
	ibne		autopilot, vi00, inautomode		; if in auto use smaller box
	lq.xyz		minvec, MinimumVecBig(vi00)	; 
	lq.xyz		maxvec, MaximumVecBig(vi00)	; 
	b			consctlpt_loop
	inautomode:
	lq.xyz		minvec, MinimumVecSmall(vi00); 
	lq.xyz		maxvec, MaximumVecSmall(vi00); 

consctlpt_loop:
	lqi.xyz		pos, (pPos++)					; load up the particle
	move.xyz	savePos, pos					; remember the save pos
	
	max.xyz		pos, pos, minvec				; constrain to box
	mini.xyz	pos, pos, maxvec				; constrain to box

	sub.xyz		vf00, savePos, pos				; if min/max changed particle 
	fmeq		vi01, allZeros					; pos, then revert the ctlpts
	ibeq		vi01, vi00, restorectlpts		; 

	isubiu		iter, iter, 1					; dec the # of particles
	ibne		iter, vi00, consctlpt_loop
	b			skiprestorectlpts	

restorectlpts:
	iaddiu		iter, vi00, 5					; for each control point move it
	iaddiu		pPos, vi00, ParticlePos			; pointer to list of particles
	iaddiu		pPosPrev, vi00, ParticlePrevPos	; pointer to particle prev pos
	iaddiu		pSaveCtlPos, vi00, SaveCtlPos	; pointer to mem to save ctlpts


restorectlpts_loop:

	lqi.xyz		pos, (pSaveCtlPos++)			; load the save pos
	sqi.xyz		pos, (pPos++)					; save to pos
	sqi.xyz		pos, (pPosPrev++)				; save to prevpos
	isubiu		iter, iter, 1					; dec the # of particles
	ibne		iter, vi00, restorectlpts_loop

skiprestorectlpts:
	
	; tug on the strings
	iaddiu		iter, vi00, 5					; for each control point move it
	iaddiu		pPos, vi00, ParticlePos			; pointer to list of particles
	iaddiu		pPosPrev, vi00, ParticlePrevPos	; pointer to particle prev pos

tugctlpt_loop:

	lq.xyz		pos, (pPos)						; load up the particle pos
	ibeq		tug0, vi00, trytug2				; if we are tugging string0
	iaddiu		string, vi00, 5					; and this is string 0
	ibne		string, iter, trytug2			
	mfir.y		ftug0, tug0						; move to a fp reg
	itof0.y		ftug0, ftug0					; make int a float
	addy.y		pos, pos, ftug0					; add y motion

	trytug2:
	ibeq		tug2, vi00, trytug3				; if we are tugging string2
	iaddiu		string, vi00, 3					; and this is string 2
	ibne		string, iter, trytug3			
	mfir.y		ftug, tug2						; move to a fp reg
	itof0.y		ftug, ftug						; make int a float
	addy.y		pos, pos, ftug					; add y motion

	trytug3:
	ibeq		tug3, vi00, trytug4				; if we are tugging string3
	iaddiu		string, vi00, 2					; and this is string 3
	ibne		string, iter, trytug4			
	mfir.y		ftug, tug3						; move to a fp reg
	itof0.y		ftug, ftug						; make int a float
	addy.y		pos, pos, ftug					; add y motion

	trytug4:
	ibeq		tug4, vi00, endtug				; if we are tugging string4
	iaddiu		string, vi00, 1					; and this is string 4
	ibne		string, iter, endtug			
	mfir.y		ftug, tug4						; move to a fp reg
	itof0.y		ftug, ftug						; make int a float
	addy.y		pos, pos, ftug					; add y motion

	endtug:
	sqi.xyz		pos, (pPos++)					; store the new pos
	sqi.xyz		pos, (pPosPrev++)				; store the new pos as prev pos

	isubiu		iter, iter, 1					; dec the # of particles
	ibne		iter, vi00, tugctlpt_loop
	

; satisfy stick and bounding box constraints
;
	; This is an iterative solution.  Loop n times
	iaddiu		iter, vi00, 3					; resolution of iteratioins

iter_loop:

	iaddiu		pSticks, vi00, StickList		; ptr to list stick constraints
	iaddiu		nSticks, vi00, NumSticks		; number of stick constraints
	iaddiu		pPos, vi00, ParticlePos			; pointer to list of particles
	iaddiu		pPosPrev, vi00, ParticlePrevPos	; pointer to particle prev pos
	iaddiu		nParticles, vi00, NumParticles	; number of particles

	iaddiu		pPos, pPos, 5
	iaddiu		pPosPrev, pPosPrev, 5
	isubiu		nParticles, nParticles, 5

	iaddiu		allZeros, vi00, AllZeroMACFlag	; the last op set all vec to 0
	lq.xyz		minvec, MinimumVecBig(vi00)		; 
	lq.xyz		maxvec, MaximumVecBig(vi00)		; 
	loi			-1.0							; used to reverse sense of sub

boundingbox_loop:

	lq.xyz		pos, (pPos)						; load up the particle
	move.xyz	savePos, pos					; remember the save pos
	
	max.xyz		pos, pos, minvec				; constrain to box
	mini.xyz	pos, pos, maxvec				; constrain to box

	sqi.xyz		pos, (pPos++)					; save the bounded particle

	sub.xyz		vf00, savePos, pos				; if min/max changed particle 
	fmeq		vi01, allZeros					; pos, then store the new pos 
	ibne		vi01, vi00, minmaxnochange		; as prev pos

	sq.xyz		pos, (pPosPrev)					; store the new pos as prev pos

minmaxnochange:

	iaddiu		pPosPrev, pPosPrev, 1			; keep pPrevPos = pPos
	isubiu		nParticles, nParticles, 1		; dec the # of particles
	ibne		nParticles, vi00, boundingbox_loop			

	iaddiu		pParticles, vi00, ParticlePos	; reset particle list pointer

stick_loop:
	ilwr.x		index1, (pSticks)x				; load the index of endpoint1
	ilwr.y		index2, (pSticks)y				; load the index of endpoint2
	lqi.z		sticklen, (pSticks++)			; load the lengh of the stick

	iadd		index1, index1, pParticles		; index into paritcle array
	iadd		index2, index2, pParticles		; index into paritcle array

	lq.xyzw		stickend1, (index1)				; load up the stick end points
	lq.xyzw		stickend2, (index2)				; load up the stick end points
	
	sub.xyz		delta, stickend2, stickend1		; diff of two points

	mul.xyz		dot, delta, delta				; find dot product of delta^2
	addy.x		dot, dot, dot
	addz.x		dot, dot, dot

	sqrt		Q, dot[x]						; sqrt (dot)

	addw.w		invmass, stickend1, stickend2	; p1.invmass + p2.invmass
	mulq.w		invmass, invmass, Q				; y = sqrt(dot) * p1.im + p2.im	
	subq.z		sticklen, sticklen, Q			; x = sticklen - sqrt(dot);
	muli.z		sticklen, sticklen, I			; x = x * -1;
	div			Q, sticklen[z], invmass[w]		; diff = x/y

	mulq.w		tmp1, stickend1, Q				; v1 = v1+delta*invmass*diff
	mulw.xyz    tmp2, delta, tmp1				;
	add.xyz		stickend1, stickend1, tmp2		;
	sq.xyz		stickend1, (index1)				; write back new point

	mulq.w		tmp1, stickend2, Q				; v2 = v2+delta*invmass*diff
	mulw.xyz    tmp2, delta, tmp1				;
	sub.xyz		stickend2, stickend2, tmp2		;
	sq.xyz		stickend2, (index2)				; write back new point
	
	isubiu		nSticks, nSticks, 1				; dec the # of sticks
	ibne		nSticks, vi00, stick_loop			
	
	isubiu		iter, iter, 1					; dec the # of iterations
	ibne		iter, vi00, iter_loop			
	
	
onlyrender:
; Render either the Skel, the meshes or both
;
	
; Render skel
;
	; for each stick constraint, render a solid line
	;
	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet
	iaddiu		pSticks, vi00, StickList		; ptr to list stick constraints
	iaddiu		nSticks, vi00, NumSticks		; number of stick constraints
	iaddiu		pParticles, vi00, ParticlePos	; pointer to list of particles

	lq			giftag, colorGifTag(vi00)		; pack the color gif tag on 
	sqi			giftag, (outpkt++)				; the output packet
	lq			giftag, green(vi00)				; pack the rgbaq for green on
	sqi			giftag, (outpkt++)				; the output packet
	
	lq			giftag, lineGifTag(vi00)		; pack the line gif tag on 
	sqi			giftag, (outpkt++)				; the output packet
	
drawlineloop:
	; --LoopCS	59,59
	
	ilwr.x		index1, (pSticks)x				; load the index of endpoint1
	ilwr.y		index2, (pSticks)y				; load the index of endpoint2
	iaddiu		pSticks, pSticks, 1				; move to next data

	iadd		index1, index1, pParticles		; index into paritcle array
	iadd		index2, index2, pParticles		; index into paritcle array

	lq.xyz		stickend1, (index1)				; load up the stick end points
	lq.xyz		stickend2, (index2)				; load up the stick end points
	
	mulax		acc, w2s0, stickend1			; transform vrt to screen space
	madday		acc, w2s1, stickend1			;
	maddaz		acc, w2s2, stickend1			;
	maddw		vrt1, w2s3, vf00				;
	
	div			q, vf00[w], vrt1[w]				; calc 1/w
	mulq.xyz 	vrt1, vrt1, q					; do perspective divide
	
	mulax		acc, w2s0, stickend2			; transform vrt to screen space
	madday		acc, w2s1, stickend2			;
	maddaz		acc, w2s2, stickend2			;
	maddw		vrt2, w2s3, vf00				;
	
	div			q, vf00[w], vrt2[w]				; calc 1/w
	mulq.xyz 	vrt2, vrt2, q					; do perspective divide
	
	ftoi4.xyz	vrt1, vrt1						; convert to fixed point
	ftoi4.xyz	vrt2, vrt2						; convert to fixed point
	sqi.xyz		vrt1, (outpkt++)				; store vrt1 into output packet
	sqi.xyz		vrt2, (outpkt++)				; store vrt2 into output packet
	
	isubiu		nSticks, nSticks, 1				; dec the number of sticks
	ibne		nSticks, vi00, drawlineloop
	
	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet
	xgkick		outpkt							; send it off
	
	iaddiu		stall, vi00, SyncGifTag			; send a NULL giftag
	xgkick		stall							; 
	
; Render meshes
;
	lq.w		l2w0, LocalWorld0(vi00)			; load local to world
	lq.w		l2w1, LocalWorld1(vi00)			; 
	lq.w		l2w2, LocalWorld2(vi00)			; 
	lq.w		l2w3, LocalWorld3(vi00)			; 
	
	iaddiu		nMeshInfo, vi00, NumMeshs		; number of times to instance
	iaddiu		pMeshInfo, vi00, MeshList		; pointer to the Mesh info
	iaddiu		pParticles, vi00, ParticlePos	; pointer to list of particles

drawmesh_loop:
	; get the local -> screen translation vector
	ilwr.x		index1, (pMeshInfo)x			; load the index of endpoint1
	ilwr.y		index2, (pMeshInfo)y			; load the index of endpoint2
	ilwr.z		index3, (pMeshInfo)z			; load the index of endpoint1
	ilwr.w		index4, (pMeshInfo)w			; load the index of endpoint2
	iaddiu		pMeshInfo, pMeshInfo, 1
	ilwr.w		useXAxis, (pMeshInfo)w
	lqi.xyz		limbscale, (pMeshInfo++)		; load the scale of the limb

	iadd		index1, index1, pParticles		; index into paritcle array
	iadd		index2, index2, pParticles		; index into paritcle array
	iadd		index3, index3, pParticles		; index into paritcle array
	iadd		index4, index4, pParticles		; index into paritcle array

	lq.xyz		limbend1, (index1)				; load up the limb end points
	lq.xyz		limbend2, (index2)				; load up the limb end points

	move.xyz	l2w3, limbend1					; add a translation to l2w

	; get the local -> screen rotation basis vectors
	sub.xyz		basis0, limbend1, limbend2		; diference of the end points
	erleng		P, basis0						; normalize
	mfp.x		flen, P							;
	mulx.xyz	basis0, basis0, flen			;

	ibne		useXAxis, vi00, useXAxis1

	lq.xyz		limbend1, (index3)				; load up the limb end points
	lq.xyz		limbend2, (index4)				; load up the limb end points
	sub.xyz		up, limbend1, limbend2			; diference of the end points
	erleng		P, up							; normalize
	mfp.x		flen, P							;
	mulx.xyz	up, up, flen					;
	b			getbasis1

	useXAxis1:
	sub.yz		up, vf00, vf00					; clear the up vector
	addw.x		up, vf00, vf00					; x-axis 1.0 0.0 0.0 0.0

	getbasis1:
	opmula.xyz	acc, basis0, up					; basis1 = up X basis0
	opmsub.xyz  basis1, up, basis0				;
	erleng		P, basis1						; normalize
	mfp.x		flen, P							;
	mulx.xyz	basis1, basis1, flen			;

	opmula.xyz	acc, basis1, basis0				; basis2 = basis0 X basis1
	opmsub.xyz	basis2, basis0, basis1			;

	move.xyz	l2w1, basis0					; populate the l2w matrix
	move.xyz	l2w0, basis2					; populate the l2w matrix
	move.xyz	l2w2, basis1					; populate the l2w matrix

	mulx.xyz	l2w0, l2w0, limbscale			; mul in the [0][0] scale
	muly.xyz	l2w1, l2w1, limbscale			; mul in the [1][1] scale
	mulz.xyz	l2w2, l2w2, limbscale			; mul in the [2][2] scale

	; build a local -> world matrix for this instance
	; [Local->Screen] = [World->Screen][Local->World]
	mulax		acc,  w2s0, l2w0
	madday		acc,  w2s1, l2w0
	maddaz		acc,  w2s2, l2w0
	maddw		l2s0, w2s3, l2w0
	mulax		acc,  w2s0, l2w1
	madday		acc,  w2s1, l2w1
	maddaz		acc,  w2s2, l2w1
	maddw		l2s1, w2s3, l2w1
	mulax		acc,  w2s0, l2w2
	madday		acc,  w2s1, l2w2
	maddaz		acc,  w2s2, l2w2
	maddw		l2s2, w2s3, l2w2
	mulax		acc,  w2s0, l2w3
	madday		acc,  w2s1, l2w3
	maddaz		acc,  w2s2, l2w3
	maddw		l2s3, w2s3, l2w3

	; build a light -> local matrix for this instance
	; [Light->Local] = [Local->World][Light]
	mulax		acc,   nl0, l2w0
	madday		acc,   nl1, l2w0
	maddaz		acc,   nl2, l2w0
	maddw		lVec0, nl3, l2w0
	mulax		acc,   nl0, l2w1
	madday		acc,   nl1, l2w1
	maddaz		acc,   nl2, l2w1
	maddw		lVec1, nl3, l2w1
	mulax		acc,   nl0, l2w2
	madday		acc,   nl1, l2w2
	maddaz		acc,   nl2, l2w2
	maddw		lVec2, nl3, l2w2
	move		lVec3, nl3

	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet

	isubiu		remain, nMeshInfo, NumBodys		; decide if we want to draw
	iblez		remain, drawbodymesh			; body mesh or limb mesh
	iaddiu		pMesh, vi00, LimbMeshData		; pointer to limb mesh data
	b			drawlimbmesh

	drawbodymesh:
	iaddiu		pMesh, vi00, BodyMeshData		; pointer to body mesh data

	drawlimbmesh:
	ilwr.x		nStrips, (pMesh)x				; number of strips

drawstrip_loop:
	ilwr.w		nVerts, (pMesh)w				; number of verts in strip
	iaddiu		pMesh, pMesh, 1					; move past control info

	lq.yzw		giftag, meshGifTag(vi00)		; read the mesh gif tag
	iaddiu		nloopeop, vi00, AllOneMask		; set to 0x8000 then and in
	iaddi		nloopeop, nloopeop, 1			; nverts
	ior			nloopeop, nloopeop, nVerts		;
	isw.x		nloopeop, (outpkt)x				; store lower 32bits of giftag
	sqi.yzw		giftag, (outpkt++)				; store high 3 32bits of giftag
	lq.xyz		basecol, fgrey200(vi00)			; load the hard coded vert color
	sub.w		basecol, vf00, vf00				; w = 0.0

	
transstripvert_loop:
	lqi.xyz		vert, (pMesh++)					; get the vert
	lqi.xyz		norm, (pMesh++)					; get the normal
	move.w		norm, vf00						; w = 1.0

	; Calc RGBAQ
	;
	mulax	 	acc, lVec0, norm				; normal x light matrix
	madday	 	acc, lVec1, norm				;
	maddaz	 	acc, lVec2, norm				;
	maddw	 	ldot, lVec3, norm				;
	maxx		ldot, ldot, vf00				; clamp

	mulax		acc, lCol0, ldot				; ldot * light color
	madday	 	acc, lCol1, ldot				;
	maddaz	 	acc, lCol2, ldot				;
	maddw		lcol, lCol3, ldot				;
	miniw		lcol, lcol, vf00

	mul			color, lcol, basecol			; calc light color * base color 
	ftoi0		color, color					; format
	sqi			color, (outpkt++)				; store color into output packet

	; Calc XYZF2
	;
	mulax		acc, l2s0, vert					; transform vert to screen space
	madday		acc, l2s1, vert					;
	maddaz		acc, l2s2, vert					;
	maddw		vert, l2s3, vf00				;
	
	div			q, vf00[w], vert[w]				; calc 1/w
	mulq.xyz 	vert, vert, q					; do perspective divide
	
	ftoi4.xy	vert, vert						; convert to fixed point
	sqi.xyz		vert, (outpkt++)				; store vert into output packet
	
	isubiu		nVerts, nVerts, 1				; dec the vert count
	ibne		nVerts, vi00, transstripvert_loop

	iaddiu  outpkt, vi00, OutputPacket			; pointer to the output packet
	xgkick	outpkt								; send it off
	
	iaddiu		stall, vi00, SyncGifTag			; send a NULL giftag
	xgkick		stall							; 
	
	isubiu		nStrips, nStrips, 1				; dec the strip count
	ibne		nStrips, vi00,drawstrip_loop

	isubiu		nMeshInfo, nMeshInfo, 1			; dec the mesh count
	ibne		nMeshInfo, vi00, drawmesh_loop

nomesh:

;;;
;;; draw the "room" that she is in
;;;

; skip all other rendering if in MATRIX mode	
	ilw.z		matrixmode, ControlQW(vi00)z	; load the doskel rendering bool
	ibne		matrixmode, vi00, noshadow		; no need to render the skel
	
	iaddiu		nStrips, vi00, 5				; nuber of strips
	iaddiu		pStrips, vi00, LineStripData	; pointer to line strip data

linelistloop1:
	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet

	lq			giftag, colorGifTag(vi00)		; pack the color gif tag on 
	sqi			giftag, (outpkt++)				; the output packet
	lq			giftag, grey(vi00)				; pack the rgbaq for grey on
	sqi			giftag, (outpkt++)				; the output packet
	
	iaddiu		nPoints, vi00, 4				; nuber of points in line

	lq.yzw		giftag, meshGifTag(vi00)		; read the mesh gif tag
	iaddiu		nloopeop, vi00, AllOneMask		; set to 0x8000 then and in
	iaddi		nloopeop, nloopeop, 1			; 
	ior			nloopeop, nloopeop, nPoints		;
	isw.x		nloopeop, (outpkt)x				; store lower 32bits of giftag
	sqi.yzw		giftag, (outpkt++)				; store high 3 32bits of giftag

	lqi.xyz		norm, (pStrips++)				; normal
	move.w		norm, vf00
	lq			basecol, fgrey128(vi00)			; pack the rgbaq for fgrey128 

linelistloop2:

	; Calc RGBAQ
	;
	mulax	 	acc, nl0, norm					; normal x light matrix
	madday	 	acc, nl1, norm					;
	maddaz	 	acc, nl2, norm					;
	maddw	 	ldot, nl3, norm					;
	maxx		ldot, ldot, vf00				; clamp

	mulax		acc, lCol0, ldot				; ldot * light color
	madday	 	acc, lCol1, ldot				;
	maddaz	 	acc, lCol2, ldot				;
	maddw		lcol, lCol3, ldot				;
	miniw		lcol, lcol, vf00

	mul			color, lcol, basecol			; calc light color * base color 
	ftoi0		color, color					; format
	sqi			color, (outpkt++)				; store color into output packet

	; Calc XYZF2
	lqi.xyz		point, (pStrips++)				; load the point

	mulax		acc, w2s0, point				; transform to screen space
	madday		acc, w2s1, point				;
	maddaz		acc, w2s2, point				;
	maddw		spoint, w2s3, vf00				;
	
	div			q, vf00[w], spoint[w]			; calc 1/w
	mulq.xyz 	spoint, spoint, q					; do perspective divide
	
	ftoi4.xyz	spoint, spoint					; convert to fixed point
	sqi.xyz		spoint, (outpkt++)				; store vrt2 into output packet

	isubiu		nPoints, nPoints, 1				; dec the point count
	ibne		nPoints, vi00, linelistloop2

	iaddiu  outpkt, vi00, OutputPacket			; pointer to the output packet
	xgkick	outpkt								; send it off
	
	iaddiu		stall, vi00, SyncGifTag			; send a NULL giftag
	xgkick		stall							; 
	
	isubiu		nStrips, nStrips, 1				; dec the point count
	ibne		nStrips, vi00, linelistloop1
	
	ilw.w		pad, LastQW(vi00)				
	isw.y		pad, ControlQW(vi00)y			; save the previous frame pad

;;;
;;; shadows
;;;

	iaddiu		nPlanes, vi00, 3				; number of planes
	iaddiu		pPlane, vi00, SaveCtlPos		; make the 3 plane normals

	sub.yzw		plane, vf00, vf00				; 1.0 0.0 0.0 0.0
	addw.x		plane, vf00, vf00				;
	loi			-44.8
	addi.w		plane, plane, i					; 1.0 0.0 0.0 -44.8
	sqi.xyzw	plane, (pPlane++)				; save the plane

	sub.xyw		plane, vf00, vf00				; 0.0 0.0 -1.0 0.0
	subw.z		plane, vf00, vf00				;
	loi			-99.8
	addi.w		plane, plane, i					; 1.0 0.0 0.0 -99.8
	sqi.xyzw	plane, (pPlane++)				; save the plane

	sub.xzw		plane, vf00, vf00				; 0.0 -1.0 0.0 0.0
	subw.y		plane, vf00, vf00				;
	loi			-31.8
	addi.w		plane, plane, i					; 0.0 1.0 0.0 -31.8
	sqi.xyzw	plane, (pPlane++)				; save the plane

	iaddiu		pPlane, vi00, SaveCtlPos		; make the 3 plane normals

	sub.xyz		light, vf00, vf00				; clear
	addx.x		light, light, nl0				; load light vector
	addx.y		light, light, nl1				;
	addx.z		light, light, nl2				;
	
shadowplane_loop:

	; create the light projection matrix (world to light - w2l)
	lqi.xyzw	plane, (pPlane++)				; load the plane normal

	mul.xyz		pdotl, light, plane				; plane dot light
	addy.x		pdotl, pdotl, pdotl
	addz.x		pdotl, pdotl, pdotl
	div			q, vf00[w], pdotl[x]			; 1/(p dot l)

	mulx.xyz	w2l0, light, plane				; Lx*Px Ly*Px Lz*Px
	subw.w		w2l0, vf00, vf00				; w = 0.0
	muly.xyz	w2l1, light, plane				; Lx*Py Ly*Py Lz*Py
	subw.w		w2l1, vf00, vf00				; w = 0.0
	mulz.xyz	w2l2, light, plane				; Lx*Pz Ly*Pz Lz*Pz
	subw.w		w2l2, vf00, vf00				; w = 0.0
	loi			-1.0
	muli.xyz	w2l0, w2l0, i					; -Lx*Px   -Ly*Px   -Lz*Px
	muli.xyz	w2l1, w2l1, i					; -Lx*Py   -Ly*Py   -Lz*Py
	muli.xyz	w2l2, w2l2, i					; -Lx*Pz   -Ly*Pz   -Lz*Pz
	mulq.xyz	w2l0, w2l0, q					; -Lx*Px*k -Ly*Px*k -Lz*Px*k
	mulq.xyz	w2l1, w2l1, q					; -Lx*Px*k -Ly*Px*k -Lz*Px*k
	mulq.xyz	w2l2, w2l2, q					; -Lx*Px*k -Ly*Px*k -Lz*Px*k
	loi			1.0
	addi.x		w2l0, w2l0, i					; 1-Lx*Px*k  -Ly*Px*k  -Lz*Px*k
	addi.y		w2l1, w2l1, i					;  -Lx*Py*k 1-Ly*Py*k  -Lz*Py*k
	addi.z		w2l2, w2l2, i					;  -Lx*Pz*k  -Ly*Pz*k 1-Lz*Pz*k
 
	mulq.xyz	w2l3, light, q					; Lx*k     Ly*k   Lz*k
	mulw.xyz	w2l3, w2l3, plane				; Lx*k*d   Ly*k*d Lz*k*d
	move.w		w2l3, vf00						; w = 1.0

; Draw line shadows
	mulax		acc,  w2s0, w2l0
	madday		acc,  w2s1, w2l0
	maddaz		acc,  w2s2, w2l0
	maddw		l2s0, w2s3, w2l0
	mulax		acc,  w2s0, w2l1
	madday		acc,  w2s1, w2l1
	maddaz		acc,  w2s2, w2l1
	maddw		l2s1, w2s3, w2l1
	mulax		acc,  w2s0, w2l2
	madday		acc,  w2s1, w2l2
	maddaz		acc,  w2s2, w2l2
	maddw		l2s2, w2s3, w2l2
	mulax		acc,  w2s0, w2l3
	madday		acc,  w2s1, w2l3
	maddaz		acc,  w2s2, w2l3
	maddw		l2s3, w2s3, w2l3

	; for each stick constraint, render a solid line
	;
	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet
	iaddiu		pSticks, vi00, StickList		; ptr to list stick constraints
	iaddiu		nSticks, vi00, NumSticks		; number of stick constraints
	iaddiu		pParticles, vi00, ParticlePos	; pointer to list of particles

	lq			giftag, colorGifTag(vi00)		; pack the color gif tag on 
	sqi			giftag, (outpkt++)				; the output packet
	;lq			giftag, green(vi00)				; pack the rgbaq for green on
	sub			giftag, vf00, vf00
	sqi			giftag, (outpkt++)				; the output packet
	
	lq			giftag, lineGifTag(vi00)		; pack the line gif tag on 
	sqi			giftag, (outpkt++)				; the output packet
	
drawlineloop2:
	; --LoopCS	59,59
	
	ilwr.x		index1, (pSticks)x				; load the index of endpoint1
	ilwr.y		index2, (pSticks)y				; load the index of endpoint2
	iaddiu		pSticks, pSticks, 1				; move to next data

	iadd		index1, index1, pParticles		; index into paritcle array
	iadd		index2, index2, pParticles		; index into paritcle array

	lq.xyz		stickend1, (index1)				; load up the stick end points
	lq.xyz		stickend2, (index2)				; load up the stick end points
	
	mulax		acc, l2s0, stickend1			; transform vrt to screen space
	madday		acc, l2s1, stickend1			;
	maddaz		acc, l2s2, stickend1			;
	maddw		vrt1, l2s3, vf00				;
	
	div			q, vf00[w], vrt1[w]				; calc 1/w
	mulq.xyz 	vrt1, vrt1, q					; do perspective divide
	
	mulax		acc, l2s0, stickend2			; transform vrt to screen space
	madday		acc, l2s1, stickend2			;
	maddaz		acc, l2s2, stickend2			;
	maddw		vrt2, l2s3, vf00				;
	
	div			q, vf00[w], vrt2[w]				; calc 1/w
	mulq.xyz 	vrt2, vrt2, q					; do perspective divide
	
	ftoi4.xyz	vrt1, vrt1						; convert to fixed point
	ftoi4.xyz	vrt2, vrt2						; convert to fixed point
	sqi.xyz		vrt1, (outpkt++)				; store vrt1 into output packet
	sqi.xyz		vrt2, (outpkt++)				; store vrt2 into output packet
	
	isubiu		nSticks, nSticks, 1				; dec the number of sticks
	ibne		nSticks, vi00, drawlineloop2
	
	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet
	xgkick		outpkt							; send it off
	
	iaddiu		stall, vi00, SyncGifTag			; send a NULL giftag
	xgkick		stall							; 
	
; Draw mesh shadows
	lq.w		l2w0, LocalWorld0(vi00)			; load local to world
	lq.w		l2w1, LocalWorld1(vi00)			; 
	lq.w		l2w2, LocalWorld2(vi00)			; 
	lq.w		l2w3, LocalWorld3(vi00)			; 
	
	iaddiu		nMeshInfo, vi00, NumMeshs		; number of times to instance
	iaddiu		pMeshInfo, vi00, MeshList		; pointer to the Mesh info
	iaddiu		pParticles, vi00, ParticlePos	; pointer to list of particles

drawmesh_loop2:
	; get the local -> screen translation vector
	ilwr.x		index1, (pMeshInfo)x			; load the index of endpoint1
	ilwr.y		index2, (pMeshInfo)y			; load the index of endpoint2
	ilwr.z		index3, (pMeshInfo)z			; load the index of endpoint1
	ilwr.w		index4, (pMeshInfo)w			; load the index of endpoint2
	iaddiu		pMeshInfo, pMeshInfo, 1
	ilwr.w		useXAxis, (pMeshInfo)w
	lqi.xyz		limbscale, (pMeshInfo++)		; load the scale of the limb

	iadd		index1, index1, pParticles		; index into paritcle array
	iadd		index2, index2, pParticles		; index into paritcle array
	iadd		index3, index3, pParticles		; index into paritcle array
	iadd		index4, index4, pParticles		; index into paritcle array

	lq.xyz		limbend1, (index1)				; load up the limb end points
	lq.xyz		limbend2, (index2)				; load up the limb end points

	move.xyz	l2w3, limbend1					; add a translation to l2w

	; get the local -> screen rotation basis vectors
	sub.xyz		basis0, limbend1, limbend2		; diference of the end points
	erleng		P, basis0						; normalize
	mfp.x		flen, P							;
	mulx.xyz	basis0, basis0, flen			;

	ibne		useXAxis, vi00, useXAxis2

	lq.xyz		limbend1, (index3)				; load up the limb end points
	lq.xyz		limbend2, (index4)				; load up the limb end points
	sub.xyz		up, limbend1, limbend2			; diference of the end points
	erleng		P, up							; normalize
	mfp.x		flen, P							;
	mulx.xyz	up, up, flen					;
	b			getbasis12

	useXAxis2:
	sub.yz		up, vf00, vf00					; clear the up vector
	addw.x		up, vf00, vf00					; x-axis 1.0 0.0 0.0 0.0

	getbasis12:
	opmula.xyz	acc, basis0, up					; basis1 = up X basis0
	opmsub.xyz  basis1, up, basis0				;
	erleng		P, basis1						; normalize
	mfp.x		flen, P							;
	mulx.xyz	basis1, basis1, flen			;

	opmula.xyz	acc, basis1, basis0				; basis2 = basis0 X basis1
	opmsub.xyz	basis2, basis0, basis1			;

	move.xyz	l2w1, basis0					; populate the l2w matrix
	move.xyz	l2w0, basis2					; populate the l2w matrix
	move.xyz	l2w2, basis1					; populate the l2w matrix

	mulx.xyz	l2w0, l2w0, limbscale			; mul in the [0][0] scale
	muly.xyz	l2w1, l2w1, limbscale			; mul in the [1][1] scale
	mulz.xyz	l2w2, l2w2, limbscale			; mul in the [2][2] scale

	; build a local -> world matrix for this instance
	; [Local->Screen] = [Light->Screen][World->Light][Local->World]
	mulax		acc,  w2l0, l2w0
	madday		acc,  w2l1, l2w0
	maddaz		acc,  w2l2, l2w0
	maddw		l2l0, w2l3, l2w0
	mulax		acc,  w2l0, l2w1
	madday		acc,  w2l1, l2w1
	maddaz		acc,  w2l2, l2w1
	maddw		l2l1, w2l3, l2w1
	mulax		acc,  w2l0, l2w2
	madday		acc,  w2l1, l2w2
	maddaz		acc,  w2l2, l2w2
	maddw		l2l2, w2l3, l2w2
	mulax		acc,  w2l0, l2w3
	madday		acc,  w2l1, l2w3
	maddaz		acc,  w2l2, l2w3
	maddw		l2l3, w2l3, l2w3

	mulax		acc,  w2s0, l2l0
	madday		acc,  w2s1, l2l0
	maddaz		acc,  w2s2, l2l0
	maddw		l2s0, w2s3, l2l0
	mulax		acc,  w2s0, l2l1
	madday		acc,  w2s1, l2l1
	maddaz		acc,  w2s2, l2l1
	maddw		l2s1, w2s3, l2l1
	mulax		acc,  w2s0, l2l2
	madday		acc,  w2s1, l2l2
	maddaz		acc,  w2s2, l2l2
	maddw		l2s2, w2s3, l2l2
	mulax		acc,  w2s0, l2l3
	madday		acc,  w2s1, l2l3
	maddaz		acc,  w2s2, l2l3
	maddw		l2s3, w2s3, l2l3

	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet

	isubiu		remain, nMeshInfo, NumBodys		; decide if we want to draw
	iblez		remain, drawbodymesh2			; body mesh or limb mesh
	iaddiu		pMesh, vi00, LimbMeshData		; pointer to limb mesh data
	b			drawlimbmesh2

	drawbodymesh2:
	iaddiu		pMesh, vi00, BodyMeshData		; pointer to body mesh data

	drawlimbmesh2:
	ilwr.x		nStrips, (pMesh)x				; number of strips

drawstrip_loop2:
	ilwr.w		nVerts, (pMesh)w				; number of verts in strip
	iaddiu		pMesh, pMesh, 1					; move past control info
	
	lq.yzw		giftag, meshGifTag(vi00)		; read the mesh gif tag
	iaddiu		nloopeop, vi00, AllOneMask		; set to 0x8000 then and in
	iaddi		nloopeop, nloopeop, 1			; nverts
	ior			nloopeop, nloopeop, nVerts		;
	isw.x		nloopeop, (outpkt)x				; store lower 32bits of giftag
	sqi.yzw		giftag, (outpkt++)				; store high 3 32bits of giftag
	
transstripvert_loop2:
	lqi.xyz		vert, (pMesh++)					; get the vert
	iaddiu		pMesh, pMesh, 1					; move past norm info

	;lq			color, grey(vi00)				; pack the rgbaq for grey on
	sub			color, vf00, vf00
	sqi			color, (outpkt++)				; store color into output packet

	; Calc XYZF2
	;
	mulax		acc, l2s0, vert					; transform vert to screen space
	madday		acc, l2s1, vert					;
	maddaz		acc, l2s2, vert					;
	maddw		vert, l2s3, vf00				;
	
	div			q, vf00[w], vert[w]				; calc 1/w
	mulq.xyz 	vert, vert, q					; do perspective divide
	
	ftoi4.xyz	vert, vert						; convert to fixed point
	sqi.xyz		vert, (outpkt++)				; store vert into output packet
	
	isubiu		nVerts, nVerts, 1				; dec the vert count
	ibne		nVerts, vi00, transstripvert_loop2

	iaddiu  	outpkt, vi00, OutputPacket		; pointer to the output packet
	xgkick		outpkt							; send it off
	
	iaddiu		stall, vi00, SyncGifTag			; send a NULL giftag
	xgkick		stall							; 
	
	isubiu		nStrips, nStrips, 1				; dec the strip count
	ibne		nStrips, vi00,drawstrip_loop2

	isubiu		nMeshInfo, nMeshInfo, 1			; dec the mesh count
	ibne		nMeshInfo, vi00, drawmesh_loop2

	isubiu		nPlanes, nPlanes, 1				; dec the plane count
	ibne		nPlanes, vi00, shadowplane_loop

noshadow:

--exit
--endexit
	
